from bolib3 import np

"""
Reference 1:
    G. Savard and J. Gauvin,
    The Steepest Descent Direction for the Nonlinear Bilevel Programming Problem
    in Operations Research Letters, Vol. 15, page 265-272, year 1994
    
Reference 2:
    Jonathan F. Bard
    Practical Bilevel Optimization: Algorithms And Applications
    Chapter 8, page 336, example 8.5.1
"""

# Properties
name: str = 'bard851'
category: str = 'textbook'
subcategory: str = 'bard_1998'
datasets: list = []

# Feasible point
x0 = np.array([1.8888888888888888])
y0 = np.array([0.8888888888888888, 0.0])


# Methods
# Upper level
def F(x, y, data=None):
    """
    Upper level objective function
    (quadratic)
    """
    return (x[0] - 1)**2 + 2*y[0]**2 - 2*x[0]


def G(x, y, data=None):
    """
    Upper level inequality constraints
    (bounds)
    x >= 0
    """
    return np.array([x[0]])


def H(x, y, data=None):
    """
    Upper level equality constraints
    (none)
    """
    return np.empty(0)


# Lower level
def f(x, y, data=None):
    """
    Lower level objective function
    (quadratic)
    """
    return (2*y[0] - 4)**2 + (2*y[1] - 1)**2 + x[0]*y[0]


Ax_lower_ineq = np.array([
    [-4.],
    [+4.],
    [-4.],
    [+4.],
    [0.],
    [0.],
])
Ay_lower_ineq = np.array([
    [-5., -4.],
    [+5., -4.],
    [+4., -5.],
    [-4., -5.],
    [1., 0.],
    [0., 1.],
])
b_lower_ineq = np.array([12., -4., 4., 4., 0, 0])


def g(x, y, data=None):
    """
    Lower level inequality constraints
    (linear)
    """
    return np.matmul(Ax_lower_ineq, x) + np.matmul(Ay_lower_ineq, y) + b_lower_ineq


def h(x, y, data=None):
    """
    Lower level equality constraints
    (none)
    """
    return np.empty(0)


def read_data(filepath=''):
    """
    If the bilevel program is parameterized by data, this function should
     provide code to read data file and return an appropriate python structure.
    """
    pass


def dimensions(key='', data=None):
    """
    If the argument 'key' is not specified, then:
     - a dictionary mapping variable/function names (str) to the corresponding dimension (int) is returned.
    If the first argument 'key' is specified, then:
     - a single integer representing the dimension of the variable/function with the name {key} is returned.
    """
    n = {
        "x": 1,  # Upper-level variables
        "y": 2,  # Lower-level variables
        "F": 1,  # Upper-level objective functions
        "G": 1,  # Upper-level inequality constraints
        "H": 0,  # Upper-level equality constraints
        "f": 1,  # Lower-level objective functions
        "g": 6,  # Lower-level inequality constraints
        "h": 0   # Lower-level equality constraints
    }
    if key in n:
        return n[key]
    return n



##################################################
# First order derivatives
##################################################
# def dF_dx(x, y):
#     """
#     First order derivative of F(x,y) with respect to x
#     """
#     return 2*x[0]
#
#
# def dF_dy(x, y):
#     """
#     First order derivative of F(x,y) with respect to y
#     """
#     return np.array([4*y[0], 0.])
#
#
# def dF(x, y):
#     """
#     First order derivative of F(x,y) with respect to x and y
#     """
#     return np.concatenate((dF_dx(x, y), dF_dy(x, y)))
#
#
# ##################################################
# # Second order derivatives
# ##################################################
# def d2F_dxx(x, y):
#     return np.array([[1.]])
#
#
# def d2F_dyy(x, y):
#     return np.array([
#         [4., 0.],
#         [0., 0.],
#     ])
#
#
# def d2F_dxy(x, y):
#     return np.array([
#         [0., 0.],
#     ])
#
#
# def d2F_dyx(x, y):
#     return np.array([
#         [0.],
#         [0.]
#     ])
#
#
# def d2F(x, y):
#     return np.block([
#         [d2F_dxx(x, y), d2F_dxy(x, y)],
#         [d2F_dyx(x, y).T, d2F_dyy(x, y)]
#     ])
